from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer

from wbintegrator_office365.models import (
    CallEvent,
    CallUser,
    Event,
    EventLog,
    Subscription,
    TenantUser,
)


class TenantUserRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbintegrator_office365:tenantuser-detail")
    profile_str = wb_serializers.StringRelatedField(source="profile")
    id_str = wb_serializers.CharField(source="id")

    class Meta:
        model = TenantUser
        fields = ("id", "tenant_id", "profile_str", "id_str", "_detail")


class SubscriptionRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbintegrator_office365:subscription-detail")

    class Meta:
        model = Subscription
        fields = ("id", "subscription_id", "created", "_detail")


class EventRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbintegrator_office365:event-detail")
    id_str = wb_serializers.CharField(source="id")

    class Meta:
        model = Event
        fields = ("id", "id_str", "uuid_event", "_detail")


class CallUserRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbintegrator_office365:calluser-detail")
    tenant_user_str = wb_serializers.StringRelatedField(source="tenant_user")
    id_str = wb_serializers.CharField(source="id")

    class Meta:
        model = CallUser
        fields = ("id", "tenant_user_str", "id_str", "_detail")


class CallEventRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbintegrator_office365:callevent-detail")

    class Meta:
        model = CallEvent
        fields = ("id", "_detail")


class TenantUserModelSerializer(wb_serializers.ModelSerializer):
    _profile = PersonRepresentationSerializer(source="profile")

    class Meta:
        model = TenantUser
        fields = (
            "id",
            "tenant_id",
            "display_name",
            "mail",
            "phone",
            "profile",
            "_profile",
            "tenant_organization_id",
            "is_internal_organization",
        )


class CallUserModelSerializer(wb_serializers.ModelSerializer):
    _tenant_user = TenantUserRepresentationSerializer(source="tenant_user")

    name_user = wb_serializers.CharField(read_only=True)
    phone = wb_serializers.CharField(read_only=True)
    mail = wb_serializers.CharField(read_only=True)

    class Meta:
        model = CallUser
        fields = read_only_fields = (
            "id",
            "tenant_user",
            "_tenant_user",
            "is_guest",
            "is_phone",
            "acs_user",
            "splool_user",
            "encrypted",
            "on_premises",
            "acs_application_instance",
            "spool_application_instance",
            "application_instance",
            "application",
            "device",
            "name_user",
            "phone",
            "mail",
        )


class SubscriptionModelSerializer(wb_serializers.ModelSerializer):
    _tenant_user = TenantUserRepresentationSerializer(source="tenant_user")
    id_str = wb_serializers.CharField(source="id")

    class Meta:
        model = Subscription
        fields = read_only_fields = (
            "id",
            "id_str",
            "subscription_id",
            "change_type",
            "notification_url",
            "resource",
            "expiration_date",
            "application_id",
            "creator_id",
            "client_state",
            "created",
            "latest_supported_tls_version",
            "notification_content_type",
            "odata_context",
            "encryption_certificate_id",
            "encryption_certificate",
            "include_resource_data",
            "notification_query_options",
            "tenant_user",
            "_tenant_user",
            "is_enable",
            "type_resource",
        )
        read_only_fields = (
            "id",
            "subscription_id",
            "change_type",
            "notification_url",
            "resource",
            "expiration_date",
            "application_id",
            "creator_id",
            "client_state",
            "created",
            "latest_supported_tls_version",
            "notification_content_type",
            "odata_context",
            "encryption_certificate_id",
            "encryption_certificate",
            "include_resource_data",
            "notification_query_options",
            "tenant_user",
            "is_enable",
            "type_resource",
        )


class SubscriptionListModelSerializer(SubscriptionModelSerializer):
    class Meta:
        model = Subscription
        fields = read_only_fields = (
            "id",
            "subscription_id",
            "change_type",
            "expiration_date",
            "type_resource",
            "resource",
            "tenant_user",
            "_tenant_user",
            "is_enable",
            "created",
        )


class EventModelSerializer(wb_serializers.ModelSerializer):
    @wb_serializers.register_resource()
    def register_history_resource(self, instance, request, user):
        resources = {
            "eventlog": reverse("wbintegrator_office365:event-eventlog-list", args=[instance.id], request=request),
        }
        return resources

    id = wb_serializers.PrimaryKeyCharField()
    id_str = wb_serializers.CharField(source="id")

    _tenant_user = TenantUserRepresentationSerializer(source="tenant_user")

    class Meta:
        model = Event
        fields = read_only_fields = (
            "id",
            "id_str",
            "auto_inc_id",
            "nb_received",
            "type",
            "subscription_id",
            "change_type",
            "resource",
            "created",
            "changed",
            "tenant_user",
            "_tenant_user",
            "_additional_resources",
            "id_event",
            "uuid_event",
            "is_handled",
        )


class EventLogRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _last_event = EventRepresentationSerializer(source="last_event")
    _detail = wb_serializers.HyperlinkField(reverse_name="wbintegrator_office365:eventlog-detail")

    class Meta:
        model = EventLog
        fields = read_only_fields = ("id", "last_event", "_last_event", "order_received", "_detail")


class EventLogModelSerializer(wb_serializers.ModelSerializer):
    _last_event = EventRepresentationSerializer(source="last_event")

    class Meta:
        model = EventLog
        fields = read_only_fields = (
            "id",
            "last_event",
            "_last_event",
            "change_type",
            "order_received",
            "created",
            "changed",
            "id_event",
            "resource",
            "is_handled",
        )


class CallEventModelSerializer(wb_serializers.ModelSerializer):
    _event = EventRepresentationSerializer(source="event")
    _organizer = CallUserRepresentationSerializer(source="organizer")
    _participants = CallUserRepresentationSerializer(source="participants", many=True)
    change_type = wb_serializers.CharField(required=False, read_only=True)

    class Meta:
        model = CallEvent
        fields = read_only_fields = (
            "id",
            "event",
            "_event",
            "organizer",
            "_organizer",
            "type",
            "change_type",
            "start",
            "end",
            "last_modified",
            "created",
            "participants",
            "_participants",
            "is_internal_call",
            "version",
            "join_web_url",
        )



from unittest.mock import patch

import pytest
from wbintegrator_office365.factories import EventFactory, SubscriptionFactory
from wbintegrator_office365.importer import DisableSignals
from wbintegrator_office365.models import (
    CallEvent,
    CallUser,
    Event,
    Subscription,
    TenantUser,
)


@pytest.mark.django_db
class TestModels:
    def test_tenantuser(self, tenant_user_factory):
        assert TenantUser.objects.count() == 0
        tenant_user_factory()
        assert TenantUser.objects.count() == 1

    def test_event(self, event_factory):
        assert Event.objects.count() == 0
        # obj = event_factory() # there is a confusion with another model event from another app
        EventFactory()
        assert Event.objects.count() == 1

    @patch("wbintegrator_office365.models.subscription.unsubscribe.delay")
    @patch("wbintegrator_office365.models.subscription.chain")
    @patch("requests.post")
    def test_subscription(self, mock_post, mock_chain, mock_unsubscribe):
        with patch("wbintegrator_office365.models.event.transaction.on_commit", new=lambda fn: fn()):
            assert Subscription.objects.count() == 0
            assert mock_chain.call_count == 0
            SubscriptionFactory()
            assert Subscription.objects.count() == 1
            assert mock_chain.call_count == 0

            assert mock_unsubscribe.call_count == 0
            SubscriptionFactory(is_enable=False)
            assert mock_chain.call_count == 1

    def test_call_event_disconnect_signal(self, call_event_factory):
        assert CallEvent.objects.count() == 0
        with DisableSignals():
            call_event_factory()
        assert CallEvent.objects.count() == 1

    def test_call_event(self, call_event_factory):
        assert CallEvent.objects.count() == 0
        call_event_factory()
        assert CallEvent.objects.count() == 1

    def test_call_user(self, call_user_factory):
        assert CallUser.objects.count() == 0
        call_user_factory()
        assert CallUser.objects.count() == 1

    def test_post_save_participant_event(self, call_event_factory, call_user_factory):
        obj = call_event_factory()
        call_user = call_user_factory()
        obj.participants.add(call_user)
        assert obj.is_internal_call is False
        obj.participants.remove(call_user)
        assert obj.is_internal_call is True



from wbintegrator_office365.viewsets.display import (
    CallEventDisplayConfig,
    EventLogDisplayConfig,
    EventLogEventDisplayConfig,
    SubscriptionDisplayConfig,
    TenantUserDisplayConfig,
)
from wbintegrator_office365.viewsets.endpoints import (
    CallEventEndpointConfig,
    EventLogEndpointConfig,
    EventLogEventEndpointConfig,
    SubscriptionEndpointConfig,
    TenantUserEndpointConfig,
)
from wbintegrator_office365.viewsets.menu import (
    CALLEVENT_MENUITEM,
    CALLEVENTCHART_MENUITEM,
    CALLEVENTSUMMARYGRAPH_MENUITEM,
    CALLUSER_MENUITEM,
    EVENT_MENUITEM,
    SUBSCRIPTION_MENUITEM,
    TENANTUSER_MENUITEM,
)
from wbintegrator_office365.viewsets.titles import (
    CallEventReceptionTimeTitleConfig,
    CallEventSummaryGraphTitleConfig,
    CallEventTitleConfig,
    CallUserTitleConfig,
    SubscriptionTitleConfig,
    TenantUserTitleConfig,
)



import random

import factory
import pytz


class TenantUserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = "wbintegrator_office365.TenantUser"

    tenant_id = factory.Sequence(lambda n: "user_tenant_%d" % n)
    display_name = factory.Faker("name")
    mail = factory.Faker("email")
    phone = factory.Faker("phone_number")
    profile = factory.SubFactory("wbcore.contrib.directory.factories.PersonFactory")
    tenant_organization_id = factory.Faker("pystr")


class SubscriptionFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = "wbintegrator_office365.Subscription"

    subscription_id = factory.Sequence(lambda n: "subscription%d" % n)
    expiration_date = factory.Faker("date_time_between", start_date="+2d", end_date="+3d", tzinfo=pytz.utc)
    change_type = factory.Faker("pystr")
    notification_url = factory.Faker("url")
    resource = factory.Faker("uri")
    application_id = factory.Faker("pystr")
    creator_id = factory.Faker("pystr")
    client_state = factory.Faker("pystr")
    latest_supported_tls_version = factory.Faker("pystr")
    notification_content_type = factory.Faker("pystr")
    odata_context = factory.Faker("pystr")
    encryption_certificate_id = factory.Faker("pystr")
    encryption_certificate = factory.Faker("pystr")
    include_resource_data = factory.Faker("pystr")
    notification_query_options = factory.Faker("pystr")
    tenant_user = factory.SubFactory(TenantUserFactory)


class EventFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = "wbintegrator_office365.Event"

    uuid_event = factory.Sequence(lambda n: "UID%d" % n)
    id_event = factory.Sequence(lambda n: "ID%d" % n)
    # type = "CALLRECORD"
    subscription_id = factory.Sequence(lambda n: "subscription%d" % n)
    # change_type = "CREATED"  #"UPDATED"
    resource = factory.Faker("uri")
    tenant_user = factory.SubFactory(TenantUserFactory)


class EventLogFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = "wbintegrator_office365.EventLog"

    last_event = factory.SubFactory(EventFactory)
    id_event = factory.Sequence(lambda n: "event%d" % n)


class CallUserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = "wbintegrator_office365.CallUser"

    acs_user = factory.Faker("pystr")
    splool_user = factory.Faker("pystr")
    encrypted = factory.Faker("pystr")
    on_premises = factory.Faker("pystr")
    acs_application_instance = factory.Faker("pystr")
    spool_application_instance = factory.Faker("pystr")
    application_instance = factory.Faker("pystr")
    application = factory.Faker("pystr")
    device = factory.Faker("pystr")
    tenant_user = factory.SubFactory(TenantUserFactory)
    is_guest = factory.Faker("pybool")
    is_phone = factory.Faker("pybool")


class CallEventFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = "wbintegrator_office365.CallEvent"
        django_get_or_create = ["activity"]

    event = factory.SubFactory(EventFactory)
    version = random.randint(1, 4)
    type = factory.Faker("pystr")
    start = factory.Faker("date_time_between", start_date="+2d", end_date="+3d", tzinfo=pytz.utc)
    end = factory.Faker("date_time_between", start_date="+2d", end_date="+3d", tzinfo=pytz.utc)
    last_modified = factory.Faker("date_time_between", start_date="+4d", end_date="+5d", tzinfo=pytz.utc)
    join_web_url = factory.Faker("url")
    organizer = factory.SubFactory(CallUserFactory)
    activity = factory.SubFactory("wbcrm.factories.ActivityFactory")
    is_internal_call = factory.Faker("pybool")

    @factory.post_generation
    def participants(self, create, extracted, **kwargs):
        if not create:
            return
        if extracted:
            for participant in extracted:
                self.participants.add(participant)



from datetime import timedelta
from unittest.mock import patch

import pytest
from rest_framework import status
from rest_framework.test import APIRequestFactory
from wbcore.contrib.authentication.factories import InternalUserFactory
from wbcore.test.utils import get_or_create_superuser
from wbhuman_resources.factories import EmployeeHumanResourceFactory
from wbintegrator_office365.factories import CallEventFactory, CallUserFactory
from wbintegrator_office365.viewsets.viewsets import (
    CallEventReceptionTime,
    CallEventSummaryGraph,
    callback_consent_permission,
    listen,
)


@pytest.mark.django_db
class TestView:
    @pytest.mark.parametrize(
        "mvs, factory",
        [
            (CallEventReceptionTime, CallEventFactory),
            (CallEventSummaryGraph, CallEventFactory),
        ],
    )
    @patch("wbintegrator_office365.importer.MicrosoftGraphAPI._get_access_token")
    @patch("wbintegrator_office365.importer.MicrosoftGraphAPI.users")
    @patch("wbintegrator_office365.importer.MicrosoftGraphAPI.get_calendar_event")
    def test_option_request(self, mock_calendar_event, mock_users, mock_acess, mvs, factory):
        request = APIRequestFactory().options("")
        request.user = get_or_create_superuser()
        factory()
        kwargs = {"user_id": request.user.id}
        vs = mvs.as_view({"options": "options"})
        response = vs(request, **kwargs)
        assert response.status_code == status.HTTP_200_OK
        assert response.data

    @pytest.mark.parametrize(
        "mvs, factory",
        [
            (CallEventReceptionTime, CallEventFactory),
            (CallEventSummaryGraph, CallEventFactory),
        ],
    )
    def test_get_plotly(self, mvs, factory):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()
        _mvs = mvs(request=request)
        fig = _mvs.get_plotly(_mvs.queryset)
        assert fig

        factory()
        _mvs = mvs(request=request)
        fig = _mvs.get_plotly(_mvs.queryset)
        assert fig

    @pytest.mark.parametrize(
        "mvs, factory, empty_compare_employee",
        [(CallEventSummaryGraph, CallEventFactory, True), (CallEventSummaryGraph, CallEventFactory, False)],
    )
    @patch("wbintegrator_office365.importer.MicrosoftGraphAPI._get_access_token")
    @patch("wbintegrator_office365.importer.MicrosoftGraphAPI.users")
    def test_get_plotly_CallEventSummaryGraph(self, mock_users, mock_acess, mvs, factory, empty_compare_employee):
        request = APIRequestFactory().get("")
        request.user = get_or_create_superuser()

        obj = factory(is_internal_call=True)
        obj.end = obj.start + timedelta(seconds=60)
        person = InternalUserFactory().profile
        employee = EmployeeHumanResourceFactory(profile=person)

        if not empty_compare_employee:
            call_user = CallUserFactory(tenant_user__profile=person)
            factory(start=obj.start, end=obj.end, type=obj.type, is_internal_call=True, participants=(call_user,))

        request.GET = request.GET.copy()
        request.GET["compare_employee"] = employee.id
        request.GET["start"] = str(obj.start.date())
        request.GET["end"] = str(obj.end.date() + timedelta(days=30))
        request.GET["call_duration"] = (obj.end - obj.start).total_seconds()
        request.GET["call_type"] = obj.type
        request.GET["call_area"] = obj.is_internal_call
        request.GET["business_day_without_call"] = "true"
        _mvs = mvs(request=request)
        fig = _mvs.get_plotly(_mvs.get_queryset())
        assert fig

    @pytest.mark.parametrize("view", [callback_consent_permission])
    def test_callback_consent_permission(self, view):
        factory = APIRequestFactory()
        request = factory.get("")
        response = view(request)
        assert response.status_code == status.HTTP_200_OK

    data = {
        "value": [
            {
                "change_type": "created",
                "subscription_id": "82507f73-75c5-4295-8ac4-5bf55f38541f",
                "resource": "communications/callRecords/af419444-a189-499c-967b-1aae574b2cc1",
                "resource_data": {"id": "af419444-a189-499c-967b-1aae574b2cc1"},
            }
        ]
    }

    @patch("wbintegrator_office365.models.event.handle_event_from_webhook.delay")
    @pytest.mark.parametrize(
        "view, validation_token, request_body",
        [(listen, None, None), (listen, "validationToken", None), (listen, None, data)],
    )
    def test_listen(self, handle_event, view, validation_token, request_body, event_factory):
        with patch("wbintegrator_office365.viewsets.viewsets.transaction.on_commit", new=lambda fn: fn()):
            factory = APIRequestFactory()
            request = factory.post("", request_body, format="json")
            if validation_token:
                request.GET = request.GET.copy()
                request.GET["validationToken"] = "fake_validation_token"
            assert handle_event.call_count == 0
            for i in range(1, 3):
                response = view(request)
                if request_body:
                    assert handle_event.call_count == i
                else:
                    assert handle_event.call_count == 0
            assert response.status_code == status.HTTP_200_OK



from django.conf import settings
from django.db import models
from wbcore.models import WBModel


class TenantUser(WBModel):
    class Meta:
        verbose_name = "Tenant User"
        verbose_name_plural = "Tenant Users"

    tenant_id = models.CharField(max_length=255, null=True, blank=True, verbose_name="User tenant ID", default="")
    display_name = models.CharField(max_length=255, null=True, blank=True, verbose_name="Display Name", default="")
    tenant_organization_id = models.CharField(
        max_length=255, null=True, blank=True, verbose_name="Organization tenant ID"
    )
    mail = models.CharField(max_length=255, null=True, blank=True, default="")
    phone = models.CharField(max_length=255, null=True, blank=True, default="")
    profile = models.ForeignKey(
        "directory.Person",
        related_name="tenant_user",
        null=True,
        blank=True,
        on_delete=models.deletion.SET_NULL,
        verbose_name="User",
    )
    is_internal_organization = models.BooleanField(null=True, blank=True)

    def save(self, *args, **kwargs):
        if self.tenant_organization_id and (org_id := getattr(settings, "WBINTEGRATOR_OFFICE365_TENANT_ID", None)):
            self.is_internal_organization = self.tenant_organization_id == org_id
        super().save(*args, **kwargs)

    def __str__(self):
        if self.profile:
            name_tenant = f"{self.profile.computed_str}"
        elif self.mail or self.display_name:
            mail = self.mail if self.mail else self.id
            name_tenant = f"{self.display_name}({mail})"
        elif self.tenant_id:
            name_tenant = f"{self.tenant_id}"
        elif self.tenant_organization_id:
            name_tenant = f"company-{self.tenant_organization_id}"
        else:
            name_tenant = f"{self.id}"
        status = "Internal" if self.is_internal_organization else "External"
        return f"{name_tenant} ({status})"

    @classmethod
    def get_endpoint_basename(cls):
        return "wbintegrator_office365:tenantuser"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbintegrator_office365:tenantuserrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{profile_str}}({{id_str}})"



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

TENANTUSER_MENUITEM = MenuItem(
    label="Tenant Users",
    endpoint="wbintegrator_office365:tenantuser-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["wbintegrator_office365.view_tenantuser", "wbintegrator_office365.administrate_event"],
    ),
)

CALLUSER_MENUITEM = MenuItem(
    label="Call Users",
    endpoint="wbintegrator_office365:calluser-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["wbintegrator_office365.view_calluser", "wbintegrator_office365.administrate_event"],
    ),
)

SUBSCRIPTION_MENUITEM = MenuItem(
    label="Subscriptions",
    endpoint="wbintegrator_office365:subscription-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["wbintegrator_office365.view_subscription", "wbintegrator_office365.administrate_event"],
    ),
)

EVENT_MENUITEM = MenuItem(
    label="Events",
    endpoint="wbintegrator_office365:event-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["wbintegrator_office365.view_event", "wbintegrator_office365.administrate_event"],
    ),
)

CALLEVENT_MENUITEM = MenuItem(
    label="Call Events",
    endpoint="wbintegrator_office365:callevent-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["wbintegrator_office365.view_callevent", "wbintegrator_office365.administrate_event"],
    ),
)

CALLEVENTCHART_MENUITEM = MenuItem(
    label="Call Reception Delay",
    endpoint="wbintegrator_office365:calleventchart-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["wbintegrator_office365.view_callevent", "wbintegrator_office365.administrate_event"],
    ),
)

CALLEVENTSUMMARYGRAPH_MENUITEM = MenuItem(
    label="Call Summary Graph",
    endpoint="wbintegrator_office365:calleventsummarygraph-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user),
        permissions=["wbintegrator_office365.view_callevent", "wbintegrator_office365.administrate_event"],
    ),
)
